unit SList;

interface

uses
  SUtil;

type
  PPointerList = ^TPointerList;
  TPointerList = array[0..MaxInt div SizeOf(Pointer)-1] of Pointer;

  TSListSortCompare = function (Item1, Item2: Pointer): Integer;
  TSList = class(TObject)
  private
    FList: PPointerList;
    FCount: Integer;
    FCapacity: Integer;
  protected
    function Get(Index: Integer): Pointer;
    procedure Grow; virtual;
    procedure Put(Index: Integer; Item: Pointer);
    procedure SetCapacity(NewCapacity: Integer);
  public
    destructor Destroy; override;
    function Add(Item: Pointer): Integer;
    procedure Clear;
    procedure Insert(Index: Integer; Item: Pointer);
    procedure Delete(Index: Integer);
    property Count: Integer read FCount;
    property Items[Index: Integer]: Pointer read Get write Put; default;
    procedure Sort(Compare: TSListSortCompare);
  end;

  TSStrList = class(TObject)
  private
    FList: TSList;
    function Get(Index: Integer): string;
    procedure Put(Index: Integer; const S: string);
    function GetCount: Integer;
  public
    constructor Create;
    destructor Destroy; override;
    function Add(const S: string): Integer;
    procedure Clear;
    property Count: Integer read GetCount;
    property Strings[Index: Integer]: string read Get write Put; default;
    procedure Sort;
  end;

  TSObjList = class(TSList)
  private
    function Get(Index: Integer): TObject;
  public
    destructor Destroy; override;
    procedure Clear;
    procedure Delete(Index: Integer);
    procedure Extract(Index: Integer);
    property Items[Index: Integer]: TObject read Get; default;
  end;

implementation

{ TSList }

destructor TSList.Destroy;
begin
  Clear;
end;

function TSList.Add(Item: Pointer): Integer;
begin
  Result := FCount;
  if Result = FCapacity then
    Grow;
  FList[Result] := Item;
  Inc(FCount);
end;

procedure TSList.Insert(Index: Integer; Item: Pointer);
begin
  if (Index < 0) or (Index > FCount) then
    Exit;
  if FCount = FCapacity then
    Grow;
  if Index < FCount then
    Move(FList[Index], FList[Index+1], (FCount-Index)*SizeOf(Pointer));
  FList[Index] := Item;
  Inc(FCount);
end;

procedure TSList.Clear;
begin
  FCount := 0;
  SetCapacity(0);
end;

procedure TSList.Delete(Index: Integer);
begin
  if (Index < 0) or (Index >= FCount) then
    Exit;
  Dec(FCount);
  if Index < FCount then
    Move(FList[Index + 1], FList[Index], (FCount - Index) * SizeOf(Pointer));
end;

function TSList.Get(Index: Integer): Pointer;
begin
  if (Index < 0) or (Index >= FCount) then
    Result := nil
  else
    Result := FList[Index];
end;

procedure TSList.Grow;
var
  d: Integer;
begin
  if FCapacity > 64 then
    d := FCapacity div 4
  else if FCapacity > 8 then
    d := 16
  else
    d := 4;
  SetCapacity(FCapacity + d);
end;

procedure TSList.Put(Index: Integer; Item: Pointer);
begin
  if (Index < 0) or (Index >= FCount) then
    Exit;
  FList[Index] := Item;
end;

procedure TSList.SetCapacity(NewCapacity: Integer);
begin
  if (NewCapacity < FCount) or (NewCapacity > (High(TPointerList)+1)) then
    Exit;
  if NewCapacity <> FCapacity then
  begin
    ReallocMem(FList, NewCapacity * SizeOf(Pointer));
    FCapacity := NewCapacity;
  end;
end;

procedure QuickSort(SortList: PPointerList; L, R: Integer;
  SCompare: TSListSortCompare);
var
  I, J: Integer;
  P, T: Pointer;
begin
  repeat
    I := L;
    J := R;
    P := SortList[(L + R) shr 1];
    repeat
      while SCompare(SortList[I], P) < 0 do Inc(I);
      while SCompare(SortList[J], P) > 0 do Dec(J);
      if I <= J then
      begin
        T := SortList[I];
        SortList[I] := SortList[J];
        SortList[J] := T;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSort(SortList, L, J, SCompare);
    L := I;
  until I >= R;
end;

procedure TSList.Sort(Compare: TSListSortCompare);
begin
  if (FList <> nil) and (FCount > 0) then
    QuickSort(FList, 0, FCount - 1, Compare);
end;


{ TSStrList }

constructor TSStrList.Create;
begin
  FList := TSList.Create;
end;

destructor TSStrList.Destroy;
begin
  Clear;
  FList.Free;
  inherited Destroy;
end;

procedure TSStrList.Clear;
var
  i: Integer;
begin
  for i := 0 to FList.Count - 1 do
    Dispose(PString(FList[i]));
  FList.Clear;
end;

function TSStrList.Get(Index: Integer): string;
begin
  Result := PString(FList[Index])^;
end;

procedure TSStrList.Put(Index: Integer; const S: string);
begin
  PString(FList[Index])^ := S;
end;

function TSStrList.GetCount: Integer;
begin
  Result := FList.Count;
end;

function TSStrList.Add(const S: string): Integer;
var
  p: PString;
begin
  New(p);
  p^ := S;
  Result := FList.Add(p);
end;

function SStrListCompare(Item1, Item2: Pointer): Integer;
begin
  Result := StrColl(PString(Item1)^, PString(Item2)^);
end;

procedure TSStrList.Sort;
begin
  FList.Sort(SStrListCompare);
end;


{ TSObjList }

destructor TSObjList.Destroy;
begin
  Clear;
  inherited Destroy;
end;

procedure TSObjList.Clear;
var
  i: Integer;
begin
  for i := 0 to Count - 1 do
    Items[i].Free;
  inherited Clear;
end;

procedure TSObjList.Extract(Index: Integer);
begin
  inherited Delete(Index);
end;

procedure TSObjList.Delete(Index: Integer);
begin
  Items[Index].Free;
  Extract(Index);
end;

function TSObjList.Get(Index: Integer): TObject;
begin
  Result := TObject(inherited Get(Index));
end;

end.


